#!/usr/bin/perl -w
# annoy.mod.pl
# -------------
#
# Copyright (C) 2001-2002 by Duke University
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
# 02111-1307, USA.
#
#
# To view the help for this module, run "perldoc template.mod.pl"
# To create the text version, run
#      "perldoc -t template.mod.pl > template.mod.txt"
# for help with POD (plain old documentation) see the "perlpod"
# man page.  POD is really easy, so check it out!
#
# $Id: annoy.mod.pl,v 1.6 2002/02/15 15:29:00 icon Exp $
#
# @Author  Konstantin Riabitsev <icon@phy.duke.edu>
# @version $Date: 2002/02/15 15:29:00 $
#

##
# Strict enforces good coding practices by making you observe the
# variable scope.
# Add other packages, too.
#
use strict;

##
# Identify ourselves (used in the mlog subroutine).
#
my $module_name = "annoy.mod";

#####################################################################
=pod

=head1 NAME

 annoy.mod.pl       module to parse out various annoyances, while
                    notifying the admin about them.
 Version            2.0, 2002-02-13
 Author             Konstantin Riabitsev <icon@phy.duke.edu>

=head1 DESCRIPTION

This module looks for various annoying things that should not be just
weeded out, such as dirty floppy mounts, gconfd locking errors, and
other misc crap.

=head1 OPTIONS

This module accepts a number of options.  All options are read from
environment variables.  These options can be set in the dulog config
file (/etc/dulog/dulog.conf by default).  Here are the options and
their default values:

=head2 GENERAL DULOG OPTIONS

These options are officially sanctioned by dulog and are common to
most modules.

=over 5

=item DEBUG => unset

When set, this module (and probably all of the others) spit debugging
info to STDOUT.

=item QUIET => unset

When set, this module (and probably all of the others) spit out a
minimal amount of info to STDOUT.  Note that this module may still
write to STDOUT if there is a serious problem that may result in log
loss.  QUIET mode overrides DEBUG mode.

=item LOGCAT => unset (read from STDIN)

This variable is set by dulog and is the file containing logs that
should be read by the module.  If unset, this module reads logs from
STDIN.

=item LOGREPORT => unset (write to STDOUT)

This variable is set by dulog and is the file to which the final
report should be written.  If unset, this module writes the report to
STDOUT.

=item LOGFILTER => unset (write to STDERR)

This variable is set by dulog and is the file to which the lines
processed by your module should be written. Dulog will fgrep this
file against the log report before mailing it (if configured to do
so). If unset, this module will output to STDERR.

=item CONFDIR => '../etc'

Location where configuration files are looked for.  NOTE: don't be
alarmed by the default value.  It's that way for easy testing and
dulog overrides it.

=back

=cut
#####################################################################

##
# General dulog options. These shouldn't change from module to module.
#
my $debug       = option('DEBUG',            undef); # dulog in debug mode
my $quiet       = option("QUIET",            undef); # dulog in quiet mode
my $input       = option("LOGCAT",           undef); # file to read
my $output      = option("LOGREPORT",        undef); # file to write report to
my $filter      = option("LOGFILTER",        undef); # file with proc'd str's
my $confdir     = option("CONFDIR",          "../etc"); # where to find config

#####################################################################
=pod

=head2 MODULE-SPECIFIC OPTIONS

These options are specific to this module.

=item $system pattern (= '%-10s -> ')

A sprintf pattern telling how to pretty-print the system name nicely.

=item $notice_pattern (= '%s (%d)')

A sprintf pattern telling how to pretty-print the note and the number
of times this activity occured.

=item $indent_pattern (= '%14s')

Simply tells sprintf how to indent the notice reports if they are coming
in for the same system.

=cut
######################################################################

##
# Sprintf rules.
#
my $system_pattern = '%-10s -> ';
my $notice_pattern = '%s (%d)';
my $indent_pattern = '%14s';

#####################################################################
=pod

=head1 CUSTOMIZING THIS MODULE

Tell people about tweaking your module here!

=cut
#####################################################################

##
# Set up logging level.
# NOTE: logs at level 0 (serious problems) are reported even when 
# "QUIET" is on!
#
my $loglevel = (defined($debug)) ? 2 : 1;
$loglevel = 0 if defined($quiet);
*LOGOUT = *STDOUT;

##
# Open the input file, $LOGCAT or read from stdin if LOGCAT isn't
# defined.
#
if (defined($input)) {
    open(INPUT, $input) or mlog(0, "cannot open input file $input\n");
} else {
    *INPUT = *STDIN;
}

##
# Open output file, $LOGREPORT or write to STDOUT if LOGREPORT
# isn't defined.
#
if (defined($output)) {
    open(OUTPUT, ">$output") or mlog(0, "cannot open output file $output\n");
} else {
    *OUTPUT = *STDOUT;
}

##
# Open the file for processed strings, or write to STDERR if 
# LOGFILTER is not defined.
#
if (defined($filter)){
    open(FILTER, ">$filter") or mlog(0, "cannot open filtered strings file\n");
} else {
    *FILTER = *STDERR;
}

##
# @filtered array is where all processed strings go.
# $struct is a "struct" variable used to hold the nested hashes of
# values.
# $flops is the same as $struct, except used solely by floppy-catcher.
#
my @filtered;
my $struct;
my $flops;

##
# Main module code.
#
mlog(1, "beginning to read input");
my $line;

##
# Iterate through all lines in the logs and analyze the ones that
# match. Analyzed and processed lines get pushed into the @filtered
# array so they can be later weeded by dulog.
#
while (defined($line = <INPUT>)) {
    my $system = systemname($line);
    ##
    # Try to find dirty floppy mounts. They usually come separated
    # in three lines. This tries to catch this garbage and organize
    # it nicely using the $flops "struct" to hold the values.
    #
    if ($line =~ /kernel: attempt to access beyond end of device/){
	push(@{$flops->{$system}}, $line);
    }
    elsif ($line =~ /kernel: .*rw=.*want=.*limit=/){
	push(@{$flops->{$system}}, $line);
    }
    elsif ($line =~ /kernel: Directory sread .* failed/){
	push(@{$flops->{$system}}, $line);
	if (defined($flops->{$system}) && $#{$flops->{$system}} == 2){
	    ##
	    # Yes, looks like a dirty floppy mount.
	    #
	    my $msg = 'dirty floppy mount';
	    addtostruct($system, $msg);
	    ##
	    # Push the lines we caught into the filtered array.
	    #
	    push(@filtered, @{$flops->{$system}});
	}
	undef $flops->{$system};
    }
    ##
    # Look for depmod messages.
    #
    elsif ($line =~ /modules.conf is more recent/){
	my $msg = 'modules.conf mismatch';
	addtostruct($system, $msg);
	push (@filtered, $line);
    }
    ##
    # Look for Gconfd locking issues. This usually occurs when
    # some user tries to log into several machines running Gnome-1.4
    # at the same time. Leaves a lot of NASTY strings in the logs.
    #
    elsif ($line =~ /gconfd.*: Failed to get lock.*Failed to create or open/
	   || $line =~ /gconfd.*: Error releasing lockfile/
	   || $line =~ /gconfd.*: .* Could not lock temporary file/
	   || $line =~ /gconfd.*: .* another process has the lock/){
	my $msg = 'gconfd locking errors';
	addtostruct($system, $msg);
	push (@filtered, $line);
    }
    ##
    # Look for fatal X errors. These usually occur when someone logs out,
    # but if they repeat a lot, then it's something that should be looked
    # at.
    #
    elsif ($line =~ /Fatal X error/){
	my $msg = 'fatal X errors';
	addtostruct($system, $msg);
	push(@filtered, $line);
    }
    ##
    # Look for sftp activity.
    #
    elsif ($line =~ /sftp-server.*:/
	   || $line =~ /subsystem request for sftp/){
	my $msg = 'sftp activity';
	addtostruct($system, $msg);
	push(@filtered, $line);
    }
    ##
    # Look for sshd misc things.
    #
    elsif ($line =~ /sshd.*: .*terminating/){
	my $msg = 'sshd terminated';
	addtostruct($system, $msg);
	push(@filtered, $line);
    }
    elsif ($line =~ /sshd.*: Server listening on/){
	my $msg = 'sshd started';
	addtostruct($system, $msg);
	push(@filtered, $line);
    }
    ##
    # Look for misc floppy errors (vmware like to leave those).
    #
    elsif ($line =~ /floppy0:/ || $line =~ /\(floppy\)/){
	my $msg = 'misc floppy errors';
	addtostruct($system, $msg);
	push(@filtered, $line);
    }
}

##
# If @filtered array is non-empty, then place the output into the
# LOGREPORT, and write the filtered strings into LOGFILTER.
#
if ($#filtered >= 0){
    mlog(1, 'Generating a report');
    print OUTPUT mkreporthdr('NOTICES');
    foreach my $system_key (sort(keys(%{$struct}))){
	my $counter = 0;
	foreach my $notice_key(keys(%{$struct->{$system_key}})){
	    $counter++;
	    my $outline;
	    if ($counter == 1){
		$outline = sprintf($system_pattern, $system_key);
	    } else {
		$outline = sprintf($indent_pattern, " ");
	    }
	    $outline .= sprintf($notice_pattern, $notice_key,
				$struct->{$system_key}{$notice_key});
	    print OUTPUT "$outline\n";
	}
    }
    mlog(1, 'Writing filtered strings');
    print FILTER join("", @filtered);
}

###################################################################
# subroutines

##
# A convenience wrapper sub which adds a hash value to the "struct"
# if it doesn't exist, or increments it by one if it does.
#
# @param  $1 The "message" used as a hash key.
# @return    void.
#
sub addtostruct {
    (my $system, my $msg) = @_;
    if (defined($struct->{$system}{$msg})){
	$struct->{$system}{$msg}++;
    } else {
	$struct->{$system}{$msg}++;
    }
}

##
# Since all syslog lines start uniformly, use this sub to
# grab the name of the system from the log line.
#
# @param  $1 The log line.
# @return    The name of the system this log line refers to.
#
sub systemname {
    my $line = shift;
    (my $system) = $line =~ m/.{15}\s(\S*)\s.*/;
    ##
    # Syslog-ng can report hosts in a more complicated way :)
    #
    if ($system =~ m{[@/](\S+)}) { $system = $1; }
    return($system);
}

##
# This sub is used for logging debug data. The logging levels are
# as follows:
# level 0 -- serious problem, should always be reported
#            (logs are probably not being reported correctly)
# level 1 -- notification of important events
# level 2 -- debugging output, useful only when coding this module
#
# Unless the log level is 0, nothing is written out to the console if
# QUIET env variable is set.
#
# @param  $1 logging level.
# @param  $2 Message to output.
# @return    void.
#
sub mlog {
    my $level = shift;
    print(LOGOUT "DULOG($module_name): ", @_, "\n") if ($level <= $loglevel);
}

##
# A wrapper to process the options passed in by environment variables.
# If the referred ENV variable is unset, then return the default value.
# This behavior is useful for debugging the module.
#
# @param  $1 The name of the ENV variable to grab.
# @param  $2 The default value to return if the ENV variable is unset.
# @return    The value of the environment variable, or the default value
#            if the variable is unset.
#
sub option {
    my $op      = shift;
    my $default = shift;
    return(exists($ENV{$op}) ? $ENV{$op} : $default);
}

##
# A wrapper to create the separator header within the report. Currently:
#
# #
# # MESSAGE
# #
#
# @param  $1 The message content.
# @return    The header formatted accordingly.
#
sub mkreporthdr {
    my $msg = shift;
    my $hdr = "#\n# $msg\n#\n";
    return $hdr;
}
