#!/usr/bin/perl -w
# logins.mod.pl
# ----------------
#
# Copyright (C) 2001-2002 by Duke University
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
# 02111-1307, USA.
#
#
# To view the help for this module, run "perldoc template.mod.pl"
# To create the text version, run
#      "perldoc -t template.mod.pl > template.mod.txt"
# For help with POD (plain old documentation) see the "perlpod"
# man page.  POD is really easy, so check it out!
#
# $Id: logins.mod.pl,v 1.14 2002/03/15 14:29:10 icon Exp $
#
# @Author  Michael D. Stenner <mstenner@phy.duke.edu>
#          Konstantin Riabitsev <icon@phy.duke.edu>
# @version $Date: 2002/03/15 14:29:10 $
#

##
# Strict enforces good coding practices by making you observe the
# variable scope.
# Add other packages, too.
#
use strict;
use Socket;

##
# Identify ourselves (used in the mlog subroutine).
#
my $module_name = "logins.mod";

#####################################################################
=pod

=head1 NAME

 logins.mod.pl   dulog module for analyzing user and damon activity
                 represented in secure.log.
 Version         2.0 (02/08/2002)
 Author          Konstantin ("Icon") Riabitsev <icon@phy.duke.edu>

=head1 DESCRIPTION

This module looks at entries in secure.log left by users logging in
and out, analyzes root logins, daemon logins, various failures, etc,
and tries to present it in convenient format so questionable activity
is noticed quickly.

=head1 OPTIONS

This module accepts a number of options.  All options are read from
environment variables.  These options can be set in the dulog config
file (/etc/dulog/dulog.conf by default).  Here are the options and
their default values:

=head2 GENERAL DULOG OPTIONS

These options are officially sanctioned by dulog and are common to
most modules.

=over 5

=item DEBUG => unset

When set, this module (and probably all of the others) spit debugging
info to STDOUT.

=item QUIET => unset

When set, this module (and probably all of the others) spit out a
minimal amount of info to STDOUT.  Note that this module may still
write to STDOUT if there is a serious problem that may result in log
loss.  QUIET mode overrides DEBUG mode.

=item LOGCAT => unset (read from STDIN)

This variable is set by dulog and is the file containing logs that
should be read by the module.  If unset, this module reads logs from
STDIN.

=item LOGREPORT => unset (write to STDOUT)

This variable is set by dulog and is the file to which the final
report should be written.  If unset, this module writes the report to
STDOUT.

=item LOGFILTER => unset (write to STDERR)

This variable is set by dulog and is the file to which the lines
processed by your module should be written. Dulog will fgrep this
file against the log report before mailing it (if configured to do
so). If unset, this module will output to STDERR.

=item CONFDIR => '../etc'

Location where configuration files are looked for.  NOTE: don't be
alarmed by the default value.  It's that way for easy testing and
dulog overrides it.

=back

=cut
#####################################################################

##
# General dulog options. These shouldn't change from module to module.
#
my $debug       = option('DEBUG',            undef); # dulog in debug mode
my $quiet       = option("QUIET",            undef); # dulog in quiet mode
my $input       = option("LOGCAT",           undef); # file to read
my $output      = option("LOGREPORT",        undef); # file to write report to
my $filter      = option("LOGFILTER",        undef); # file with proc'd str's
my $confdir     = option("CONFDIR",          "../etc"); # where to find config

#####################################################################
=pod

=head2 MODULE-SPECIFIC OPTIONS

You may configure the following options to change how the report looks
like.

=item $report_pattern (= '%-10s-> %10s: ')

This variable specifies a sprintf string to use when beginning each
host/user report. Currently it will provide for root:  "hostname  
->  service: ", for users: "username  ->  service: ".

=item $next_line_pattern (= '%23s: ')

This variable defines how the next line should be indented, should there
be more than one service reported per host/username.

=item $sysindent_pattern (= '%23s  ')

This variable defines the amount of indentation used when wrapping
long lines.

=item $sysreport_pattern (= '%s(%d) ')

This variable defines how the right-side formatting of a report is
done.  E.g. currently it is for root: "byuser@rhost(times) ", for
users: "hostname(times) "

=item $notice_pattern (= '%s from %s: %d times')

This variable defines how the notices are formatted. E.g.: 
"SSH version scan from malicious.hacker.tld: 200 times".

=item $wrap

Hard limit at which to wrap the lines. The wrapped lines will be indented
according to the $sysindent_pattern variable.

=cut
#####################################################################

##
# Formatting options
#
my $report_pattern      = '%-10s-> %10s: ';
my $next_line_pattern   = '%23s: ';
my $sysindent_pattern   = '%23s  ';
my $sysreport_pattern   = '%s(%d) ';
my $notice_pattern      = '%s from %s: %d times';
my $wrap = 76;

#####################################################################
=pod

=head1 CUSTOMIZING THIS MODULE

Tell people about tweaking your module here!

=cut
#####################################################################

##
# Set up logging level.
# NOTE: logs at level 0 (serious problems) are reported even when 
# "QUIET" is on!
#
my $loglevel = (defined($debug)) ? 2 : 1;
$loglevel = 0 if defined($quiet);
*LOGOUT = *STDOUT;

##
# Open the input file, $LOGCAT or read from stdin if LOGCAT isn't
# defined.
#
if (defined($input)) {
    open(INPUT, $input) or mlog(0, "cannot open input file $input\n");
} else {
    *INPUT = *STDIN;
}

##
# Open output file, $LOGREPORT or write to STDOUT if LOGREPORT
# isn't defined.
#
if (defined($output)) {
    open(OUTPUT, ">$output") or mlog(0, "cannot open output file $output\n");
} else {
    *OUTPUT = *STDOUT;
}

##
# Open the file for processed strings, or write to STDERR if 
# LOGFILTER is not defined.
#
if (defined($filter)){
    open(FILTER, ">$filter") or mlog(0, "cannot open filtered strings file\n");
} else {
    *FILTER = *STDERR;
}

##
# Define any global variables here.
# @filtered array is where all processed strings go.
#
my @filtered;
##
# These hashes are used to cache the hosts and uids.
#
my %known_hosts;
my %known_uids;
##
# Define some constants as the root hash elements of the struct.
# Root's constants are 10 more than users' so we can just add 10
# to it and arrive to roots' when we do the parsing.
#
my $ACTION_FAILURE      = 0;
my $ACTION_OPENED       = 1;
my $ACTION_CLOSED       = 2;
my $ACTION_ROOT_FAILURE = 10;
my $ACTION_ROOT_OPENED  = 11;
my $ACTION_ROOT_CLOSED  = 12;
my $ACTION_IGNORE       = 3;
my $ACTION_NOTICE       = 4;
##
# This is a "struct" that holds nested hashes.
#
my $struct;

##
# Do the magic.
#
mlog(1, "beginning to read input");
##
# Process the lines first
#
my $line;
while (defined($line = <INPUT>)) {
    my @ry = processLine($line);
    ##
    # $ry[0] will be undefined if the line wasn't caught by the
    # processLine subroutine.
    #
    if (defined($ry[0])){
	##
	# Add it to the @filtered array, so it can be later weeded out
	# by dulog.
	#
	push(@filtered, $line);
	##
	# Make sure it's not ACTION_IGNORE
	#
	if ($ry[0] != $ACTION_IGNORE){
	    if (defined($ry[3]) && $ry[3] eq "root"){
		##
		# root user gets a different layout.
		# we don't care about xscreensaver, because it will always
		# scream as if someone is trying to log in as root.
		#
		if ($ry[2] ne "xscreensaver"){
		    if (!defined($ry[4])){
			$ry[4] = "";
		    }
		    if (!defined($ry[5])){
			$ry[5] = "";
		    }
		    my $ruhost = $ry[4].'@'.$ry[5];
		    $ruhost =~ s/\@$//;
		    if ($ruhost eq ""){
			$ruhost = 'unknown';
		    }
		    if (!defined($struct->{$ry[0]}{'root'}{$ry[1]}{$ry[2]}->
				 {$ruhost})){
			$struct->{$ry[0]}{'root'}{$ry[1]}{$ry[2]}{$ruhost}=1;
		    } else {
			$struct->{$ry[0]}{'root'}{$ry[1]}{$ry[2]}{$ruhost}++;
		    }
		}
	    } else {
		if ($ry[0] == $ACTION_NOTICE){
		    ##
		    # notices get special treatment
		    #
		    if (defined($struct->{$ACTION_NOTICE}{$ry[1]}{$ry[2]})){
			$struct->{$ACTION_NOTICE}{$ry[1]}{$ry[2]}++;
		    } else {
			$struct->{$ACTION_NOTICE}{$ry[1]}{$ry[2]}=1;
		    }		
		} else {
		    ##
		    # general users
		    #
		    if (!defined($struct->{$ry[0]}{$ry[3]}{$ry[2]}{$ry[1]})){
			$struct->{$ry[0]}{$ry[3]}{$ry[2]}{$ry[1]}=1;
		    } else {
			$struct->{$ry[0]}{$ry[3]}{$ry[2]}{$ry[1]}++;
		    }
		}
	    }
	}
    }
}
mlog(1, "Finished reading input");

if ($#filtered >= 0){
    ##
    # Process root failures.
    #
    if (defined($struct->{$ACTION_ROOT_FAILURE})){
	mlog(1, "Writing root login failures");
	print OUTPUT mkreporthdr('ROOT LOGIN FAILURES');
	outputRootEntries($ACTION_ROOT_FAILURE);
    }

    ##
    # Process root successes
    #
    if (defined($struct->{$ACTION_ROOT_OPENED})){
	mlog(1, "Writing root logins");
	print OUTPUT mkreporthdr('ROOT LOGINS');
	outputRootEntries($ACTION_ROOT_OPENED);
    }

    ##
    # Process user failures.
    #
    if (defined($struct->{$ACTION_FAILURE})){
	mlog(1, "Writing user login failures");
	print OUTPUT mkreporthdr('User login FAILURES');
	outputUserEntries($ACTION_FAILURE);
    }

    ##
    # Process user successes.
    #
    if (defined($struct->{$ACTION_OPENED})){
	mlog(1, "Writing user logins");
	print OUTPUT mkreporthdr('User logins');
	outputUserEntries($ACTION_OPENED);
    }

    ##
    # Process notices.
    #
    if (defined($struct->{$ACTION_NOTICE})){
	mlog(1, "Writing notices");
	print OUTPUT mkreporthdr('Login module notes');
	foreach my $descr_key (keys(%{$struct->{$ACTION_NOTICE}})){
	    foreach my $rhost_key (keys(%{$struct->{$ACTION_NOTICE}->
					      {$descr_key}})){
		my $outline = sprintf ($notice_pattern, $descr_key, $rhost_key,
				       $struct->{$ACTION_NOTICE}{$descr_key}->
				       {$rhost_key});
		print OUTPUT "$outline\n";
	    }
	}
    }
    ##
    # Write processed strings to the file, if we processed any.
    #
    print FILTER join("", @filtered);
}


###################################################################
# subs and salads

##
# This sub wraps the lines so they are $wrap width max.
#
# @param  $1 The string to wrap.
# @return    The same string cleanly wrapped.
#
sub cleanWrap {
    (my $lline) = @_;
    my @retlines;
    if (length($lline)>$wrap){
	while (length($lline)>0){
	    (my $str, my $rest) = $lline =~ m/^(.{0,$wrap})\s(.*)/;
	    push(@retlines, $str);
	    if (length($rest)>0){
		$lline = sprintf($sysindent_pattern, " ") . $rest;
	    } else {
		$lline = "";
	    }
	}
	return join("\n", @retlines);
    } else {
	return $lline;
    }
}

##
# This sub outputs noticed root entries in a nice format.
#
# @param  $1 the root element of the struct to process.
# @return    void.
#
sub outputRootEntries {
    (my $ACT) = @_;
    foreach my $system_key (sort(keys(%{$struct->{$ACT}{'root'}}))){
	my $counter = 0;
	foreach my $service_key (keys(%{$struct->{$ACT}{'root'}->
					    {$system_key}})){
	    $counter++;
	    my $longline;
	    if ($counter == 1){
		$longline = sprintf($report_pattern, $system_key, 
				    $service_key);
	    } else {
		$longline = sprintf($next_line_pattern, $service_key);
	    }
	    foreach my $ruhost_key (keys(%{$struct->{$ACT}->
					       {'root'}{$system_key}->
						   {$service_key}})){
		$longline .= sprintf($sysreport_pattern, $ruhost_key,
				     $struct->{$ACT}{'root'}->
				     {$system_key}{$service_key}->
				     {$ruhost_key});
	    }
	    print OUTPUT cleanWrap($longline) . "\n";
	}
    }
}

##
# This sub outputs user entries in a nice format.
#
# @param  $1 The root element of the struct to process.
# @return    void.
#
sub outputUserEntries {
    (my $ACT) = @_;
    foreach my $user_key (sort(keys(%{$struct->{$ACT}}))){
	my $counter = 0;
	foreach my $service_key (keys(%{$struct->{$ACT}{$user_key}})){
	    $counter++;
	    my $longline;
	    if ($counter == 1){
		$longline = sprintf($report_pattern, $user_key,
				    $service_key);
	    } else {
		$longline = sprintf($next_line_pattern, $service_key);
	    }
	    foreach my $system_key (sort(keys(%{$struct->{$ACT}->
						    {$user_key}->
							{$service_key}}))){
		$longline .= sprintf($sysreport_pattern, $system_key,
				     $struct->{$ACT}{$user_key}->
				     {$service_key}{$system_key});
	    }
	    print OUTPUT cleanWrap($longline) . "\n";
	}
    }
}

##
# This sub does all the processing. It's long and painful. ;)
#
# @param  $1 takes a log line.
# @return    an array of parameters. They vary depending on whether it's
#            a root-user, common user, or a notice.
#
sub processLine {
    my $action;
    my $system;
    my $service;
    my $user;
    my $byuser;
    my $rhost;
    my $descr;
    (my $line) = @_;
    $system = systemname($line);

    ##
    # Analyze per-report style.
    #

    ################## pam_unix #################
    if ($line =~ /(pam_unix)/){
	##
	# the line we're looking for would be service(pam_unix)
	#
	($service) = $line =~ m/.*\s(\S*)\(pam_unix\).*/;
	if ($line =~ /authentication failure/){
	    $action = $ACTION_FAILURE;
	    ($user) = $line =~ m/.*\suser=(\S*).*/;
	    ($byuser) = $line =~ m/.*\slogname=(\S*).*/;
	    ($rhost) = $line =~ m/.*\srhost=(\S*).*/;
	}
	elsif ($service eq "sshd"){
	    ##
	    # ignore sshd closes and successes
	    #
	    $action = $ACTION_IGNORE;
	} else {
	    if ($line =~ /session opened/){
		$action = $ACTION_OPENED;
		($user) = $line =~ m/.* for user (\S*).*/;
		($byuser) = $line =~ m/.* by.*\(uid=(\S*)\).*/;
		$byuser = username($byuser);
	    }
	    elsif ($line =~ /session closed/){
		$action = $ACTION_CLOSED;
		($user) = $line =~ m/.* for user (\S*).*/;
	    }
	    ##
	    # misc failure lines
	    #
	    elsif ($line =~ /bad username/){
		$action = $ACTION_FAILURE;
		($user) = $line =~ m/.*\[(.*)\].*/;
	    }
	    elsif ($line =~ /auth could not identify password for/){
		$action = $ACTION_FAILURE;
		($user) = $line =~ m/.*\[(.*)\].*/;
	    }
	    elsif ($line =~ /check pass;/){
		$action = $ACTION_FAILURE;
		($user) = $line =~ m/.* user (\S*).*/;
	    }
	}
    }
    ######################## xinetd ################
    elsif ($line =~ /xinetd\[\S*\]:/){
	($service) = $line =~ m/.* xinetd\[\S*\]: \S*: (\S*) .*/;
	if ($line =~ /START:/){
	    $action = $ACTION_OPENED;
	    ($rhost) = $line =~ m/.* from=(\S*).*/;
	    $rhost = hostname($rhost);
	}
	elsif ($line =~ /EXIT:/){
	    $action = $ACTION_CLOSED;
	}
	##
	# catch and ignore imap, since we process separate
	# logins from them.
	#
	elsif ($line =~ / imap/){
	    $action = $ACTION_IGNORE;
	}
    }
    ######################## PAM_unix ####################
    elsif ($line =~ /PAM_unix\[/){
	($service) = $line =~ m/.*PAM_unix\[\S*\]: \((\S*)\) .*/;
	if ($line =~ /authentication failure/){
	    $action = $ACTION_FAILURE;
	    ($service) = $line =~ m/.* for (\S*) service.*/;
	    ($user) = $line =~ m/.*-> (\S*) .*/;
	    ($byuser) = $line =~ m/.* failure; \(uid=(\S*)\) .*/;
	    $byuser = username($byuser);
	}
	##
	# ignore sshd successes and closes, since we do them later
	#
	elsif ($service eq "sshd") {
	    $action = $ACTION_IGNORE;
	} else {
	    if ($line =~ /session opened/){
		$action = $ACTION_OPENED;
		($user) = $line =~ m/.* for user (\S*) .*/;
		($byuser) = $line =~ m/.* by \(uid=(\S*)\).*/;
		$byuser = username($byuser);
	    }
	    elsif ($line =~ /session closed/){
		$action = $ACTION_CLOSED;
		($user) = $line =~ m/.* for user (\S*).*/;
	    }
	}
    }
    ##
    # PAM_unix and PAM_pwdb are nearly identical, but not entirely.
    #

    ######################## PAM_pwdb ####################
    elsif ($line =~ /PAM_pwdb\[/){
	if ($line =~ /authentication failure/){
	    $action = $ACTION_FAILURE;
	    ($service) = $line =~ m/.* for (\S*) service.*/;
	    ($user) = $line =~ m/.*-> (\S*) .*/;
	    ($byuser) = $line =~ m/.* failure; \(uid=(\S*)\) .*/;
	    $byuser = username($byuser);
	}
	##
	# ignore sshd, since we'll process it separately
	#
	elsif ($line =~ /\(sshd\)/){
	    $action = $ACTION_IGNORE;
	} else {
	    if ($line =~ m/.*PAM_pwdb\[\S*\]: \((\S*)\) .*/){
		($service) = $line =~ m/.*PAM_pwdb\[\S*\]: \((\S*)\) .*/;
	    } else {
		$service = 'unknown';
	    }
	    if ($line =~ /session opened/){
		$action = $ACTION_OPENED;
		($user) = $line =~ m/.* for user (\S*) .*/;
		($byuser) = $line =~ m/.*\(uid=(\S*)\).*/;
		$byuser = username($byuser);
	    }
	    elsif ($line =~ /session closed/){
		$action = $ACTION_CLOSED;
		($user) = $line =~ m/.* for user (\S*).*/;
	    }
	}
    }
    ########################### sshd #########################
    elsif ($line =~ /sshd\[/){
	if ($line =~ /ssh2/){
	    $service = "ssh2";
	} else {
	    $service = "ssh1";
	}
	if ($line =~ /: Accepted/){
	    $action = $ACTION_OPENED;
	    ($user) = $line =~ m/.* for (\S*) from.*/;
	    ($byuser) = $line =~ m/.* ruser (\S).*/;
	    ($rhost) = $line =~ m/.* from (\d\S*)/;
	    $rhost = hostname($rhost);
	    (my $method) = $line =~ m/.*: Accepted (\S*) for.*/;
	    if ($method eq "password"){
		$service .= "(pw)";
	    } elsif ($method eq "publickey"){
		$service .= "(pk)";
	    } elsif ($method eq "rhosts-rsa" || $method eq "rsa"){
		$service .= "(rsa)";
	    }
	}
	elsif ($line =~ /: Connection closed/){
	    $action = $ACTION_CLOSED;
	    ($rhost) = $line =~ m/.* closed by (\S*).*/;
	    ##
	    # don't do this, as we are currently ignoring the closes.
	    #
	    #$rhost = hostname($rhost);
	}
	elsif ($line =~ /: Failed/){
	    ##
	    # ignore these, since pam_unix catches these better
	    #
	    $action = $ACTION_IGNORE;
	}
	elsif ($line =~ /: Rhosts authentication refused/){
	    $action = $ACTION_FAILURE;
	    $service = "ssh(rsa)";
	    ($user) = $line =~ m/.*refused for (\S*):.*/;
	}
	##
	# recognize and add some notes
	#
	elsif ($line =~ /Did not receive ident/){
	    $action = $ACTION_NOTICE;
	    ($rhost) = $line =~ m/.*string from (\d\S*).*/;
	    $rhost = hostname($rhost);
	    $descr = "SSH connect scan";
	}
	elsif ($line =~ /Version_Mapper/){
	    $action = $ACTION_IGNORE;
	    ($rhost) = $line =~ m/.*scanned from (\d\S*).*/;
	    $rhost = hostname($rhost);
	    $descr = "SSH version mapper scan";
	}
	##
	# Some silliness with ROOT vs root.
	#
	if (defined($user) && $user eq "ROOT"){
	    $user = "root";
	}
    }
    ############# imapd ##############
    elsif ($line =~ /imapd\[\d*\]:/){
	$service = 'imap';
	($user) = $line =~ m/.* user=(\S*).*/;
	($rhost)= $line =~ m/.*host=(\S*).*/;
	if (!defined($rhost)){
	    ($rhost) = $line =~ m/.*\[\S*\].*/;
	    $rhost = hostname($rhost);
	}
	if ($line =~ /Authenticated user/){
	    $action = $ACTION_OPENED;
	}
	elsif ($line =~ /Logout/){
	    $action = $ACTION_CLOSED;
	}
	elsif ($line =~ /Killed/){
	    $action = $ACTION_CLOSED;
	}
	elsif ($line =~ /Login failure/){
	    $action = $ACTION_FAILURE;
	}
	elsif ($line =~ /AUTHENTICATE LOGIN/ ||
	       $line =~ /AUTHENTICATE PLAIN/){
	    ##
	    # ignore these since they just duplicate the ones we
	    # already caught above.
	    #
	    $action = $ACTION_IGNORE;
	}
    }
    ##
    # Catch and ignore xscreensaver, since pam_unix catches it.
    #
    elsif ($line =~ /xscreensaver\[\d*\]: FAILED LOGIN/){
	$action = $ACTION_IGNORE;
    }

    if (defined($action)){
	if ($action == $ACTION_NOTICE){
	    return ($action, $descr, $rhost);
	} else {
	    if (!defined($user)){
		$user = "unknown";
	    }
	    if ($user eq "root"){
		##
		# up it by 10 to get root-level actions.
		#
		$action += 10;
	    }
	    return ($action, $system, $service, $user, $byuser, $rhost);
	}
    } else {
	return (undef);
    }
}

##
# This sub takes a uid and looks up the user name.
#
# @param  $1 the uid to look up.
# @return    the username.
#
sub username {
    my $uid = shift;
    if (!defined($uid)){
	return(undef);
    }

    if (exists($known_uids{$uid})){
	return($known_uids{$uid});
    } else {
	(my $uname) = getpwuid($uid);
	if (!defined($uname)){
	    $uname = "uid=$uid";
	}
	$known_uids{$uid} = $uname;
	return($uname);
    }
}

##
# A wrapper to create the separator header within the report. Currently:
#
# #
# # MESSAGE
# #
#
# @param  $1 The message content.
# @return    The header formatted accordingly.
#
sub mkreporthdr {
    my $msg = shift;
    my $hdr = "#\n# $msg\n#\n";
    return $hdr;
}

##
# Since all syslog lines start uniformly, use this sub to
# grab the name of the system from the log line.
#
# @param  $1 The log line.
# @return    The name of the system this log line refers to.
#
sub systemname {
    my $line = shift;
    (my $system) = $line =~ m/.{15}\s(\S*)\s.*/;
    ##
    # syslog-ng can report hosts in a more complicated way :)
    #
    if ($system =~ m{[@/](\S+)}) { $system = $1; }
    return($system);
}

### Michael Stenner's subroutines ###

##
# This sub is used for logging debug data. The logging levels are
# as follows:
# level 0 -- serious problem, should always be reported
#            (logs are probably not being reported correctly)
# level 1 -- notification of important events
# level 2 -- debugging output, useful only when coding this module
#
# Unless the log level is 0, nothing is written out to the console if
# QUIET env variable is set.
#
# @param  $1 logging level.
# @param  $2 Message to output.
# @return    void.
#
sub mlog {
    my $level = shift;
    print(LOGOUT "DULOG($module_name): ", @_, "\n") if ($level <= $loglevel);
}

##
# A wrapper to process the options passed in by environment variables.
# If the referred ENV variable is unset, then return the default value.
# This behavior is useful for debugging the module.
#
# @param  $1 The name of the ENV variable to grab.
# @param  $2 The default value to return if the ENV variable is unset.
# @return    The value of the environment variable, or the default value
#            if the variable is unset.
#
sub option {
    my $op      = shift;
    my $default = shift;
    return(exists($ENV{$op}) ? $ENV{$op} : $default);
}

##
# This sub tries to resolve hostnames if possible. If not, it returns
# the ip address back. The %known_hosts hash is used to cache the values
# for optimization.
#
# @param  $1 The IP of a host to lookup.
# @return    The FQDN, or the IP address if lookup failed.
#
sub hostname {
    my $host = shift;
    if (exists($known_hosts{$host})) {
	return($known_hosts{$host});
    } else {
	##
	# hash resolved names.  This speeds things up because we often get
	# many hits from the same host.
	#
	my @host_a = gethostbyaddr(pack('CCCC', split(/\./, $host)), AF_INET);
	my $hostname = defined($host_a[0]) ? $host_a[0] : $host;
	$known_hosts{$host} = $hostname;
	return($hostname);
    }
}
